!---------------------------------------------------------------------------------------
! FLEXINVERT: init_co2
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> init_co2
!! Initialize CO2 variables
!! 
!! Purpose:    Reads prior fluxes and observations of CO2 and calculates the 
!!             state and observation vectors as well as their uncertainties.
!!
!! Interface:
!!
!!    Inputs
!!             files   -  file data structure
!!             config  -  configuration settings data structure
!!             fluxes  -  flux data structure
!!             obs     -  observation data structure
!!             covar   -  error covariance data structure
!!
!!    Externals
!!             alloc_fluxes
!!             alloc_states
!!             caldate     
!!             read_ncdf        
!!             read_bg       
!!             correl
!!             error_cov        
!!             read_obs
!!             assign_obs
!!             init_cini
!!             perturb_state
!!             perturb_obs
!!
!---------------------------------------------------------------------------------------

subroutine init_co2(files, config, fluxes, obs, states, covar)

  use mod_var
  use mod_settings
  use mod_strings
  use mod_dates
  use mod_tools
  use mod_fluxes
  use mod_obs
  use mod_states
  use mod_covar
  use mod_perturb
  use mod_save, only : save_for_restart

  implicit none

  type (files_t),         intent (in out)                :: files
  type (config_t),        intent (in)                    :: config
  type (fluxes_t),        intent (in out)                :: fluxes
  type (obs_t),           intent (in out)                :: obs
  type (states_t),        intent (in out)                :: states
  type (covar_t),         intent (in out)                :: covar

  character(max_path_len)                                :: filename
  character(max_path_len)                                :: rowfmt
  character(len=4)                                       :: adate
  character(len=2)                                       :: amonth
  logical                                                :: lexist
  real, dimension(nbox,ntstate,ndt)                      :: err_box
  real, dimension(nbox,ntstate,nd_nee)                   :: nee_box
  real, dimension(nbox,nbox)                             :: corr
  real(kind=8), dimension(:), allocatable                :: timefp
  real, dimension(nxregrid,nyregrid,nd_nee)              :: reg_nee
  real(kind=8), dimension(:), allocatable                :: timenee, timeff, timeocn
  real, dimension(:,:,:), allocatable                    :: flx, flxnee, flxff, flxocn
  real, dimension(:), allocatable                        :: xerr
  integer, dimension(ndt)                                :: cnt
  integer                                                :: nread, num, numocn
  integer                                                :: eomday, eomday_save
  integer                                                :: jjjjmmdd, hhmiss, hh
  real                                                   :: area
  real(kind=8)                                           :: jd, jdi 
  integer                                                :: ix, jy, ix0, jy0
  integer                                                :: n, nb, nd, nt, nts, i
  integer                                                :: ierr, lenstr
  real(kind=8), dimension(:), allocatable                :: jdate
  character(len=recname_len), dimension(:), allocatable  :: recs
  real                                                   :: conc
  character(len=20)                                      :: dump
  real, dimension(:,:), allocatable                      :: datain, dataout
  real, dimension(nxregrid,nyregrid)                     :: tmp

  ! global prior fluxes
  ! -------------------

  ! global fluxes are used for calculating contribution to mixing ratios
  ! from outside the inversion domain i.e. background and are averaged to 
  ! the temporal resolution of the state vector

  ! note: for global inversions the fossil fuel and NEE fluxes 
  ! averaged to the state vector temporal resolution are not needed
  ! (no background contribution from fluxes) but are calculated any how to
  ! avoid additional exceptions in the code

  call alloc_fluxes(config, fluxes)
  fluxes%flxnee(:,:,:) = 0.
  fluxes%flxff(:,:,:) = 0.
  fluxes%flxocn(:,:,:) = 0.
  fluxes%time(:) = 0.

  ! loop over global prior flux files

  n = 0
  jd = juldatei

  do while ( (jd.le.juldatef).and.(n.lt.ntstate) )

    n = n + 1
    call caldate(jd, jjjjmmdd, hhmiss)
    write(adate,'(I4.4)') jjjjmmdd/10000
    write(amonth,'(I2.2)') (jjjjmmdd-(jjjjmmdd/10000)*10000)/100

    !--- NEE fluxes

    ! number of fields to read for each flux time step
    nread = int(statres)*24/nt_nee
    print*, 'init_co2: nee nread = ',nread
    allocate( flx(nxgrid,nygrid,nread), stat = ierr )
    if ( ierr.ne.0 ) stop 'ERROR init_co2: not enough memory'
    filename = str_replace(files%filename_nee, 'YYYY', adate)
    if ( index(filename, 'MM').ne.0 ) then
      filename = str_replace(filename, 'MM', amonth)
    endif
    filename = trim(files%path_prior)//trim(filename)
    inquire(file=trim(filename),exist=lexist)
    if (.not.lexist) then    
      write(logid,*) 'ERROR: cannot find '//trim(filename)
      stop
    endif
    write(logid,*) 'Reading NEE :'//trim(filename)
    call read_ncdf(filename, &
                   files%varname_nee, &
                   files%lonname_nee, &
                   files%latname_nee, &
                   files%timename_nee,&
                   llx, lly, &
                   nxgrid, nygrid, &
                   jd, nread, num, &
                   flx)
    print*, 'init_co2: nee num = ',num
    ! convert to kg/m2/s
    flx = flx/3600.
    ! apply numerical scaling
    flx = flx*numscale
    ! average field to flux time step 
    fluxes%flxnee(:,:,n) = sum(flx,dim=3)/real(nread)
    deallocate( flx )

    !--- fossil fuel fluxes

    ! number of fields to read for each flux time step
    nread = int(statres*24./real(nt_ff))
    if ( nread.eq.0 ) nread = 1
    print*, 'init_co2: ff nread = ',nread
    allocate( flx(nxgrid,nygrid,nread), stat = ierr )
    if ( ierr.ne.0 ) stop 'ERROR init_co2: not enough memory'
    filename = str_replace(files%filename_ff, 'YYYY', adate)
    if ( index(filename, 'MM').ne.0 ) then
      filename = str_replace(filename, 'MM', amonth)
    endif
    filename = trim(files%path_prior)//trim(filename)
    inquire(file=trim(filename),exist=lexist)
    if (.not.lexist) then    
      write(logid,*) 'ERROR: cannot find '//trim(filename)
      stop
    endif
    write(logid,*) 'Reading fossil fuel :'//trim(filename)
    if ( nt_ff.lt.24 ) then
      ! assume timestamp is hours since 1-Jan each year
      jdi = (jd - juldate((jjjjmmdd/10000)*10000+101,0))*24d0 + juldate(19000101,0)
    else
      jdi = jd
    endif
    call read_ncdf(filename, &
                   files%varname_ff, &
                   files%lonname_ff, &
                   files%latname_ff, &
                   files%timename_ff,&
                   llx, lly, &
                   nxgrid, nygrid,  &
                   jdi, nread, num, &
                   flx)
    print*, 'init_co2: ff num = ',num
    ! convert from input flux units to kg/m2/s
    flx = flx*config%coeff_ff/3600.
    ! apply numerical scaling
    flx = flx*numscale
    ! average field to flux time step 
    fluxes%flxff(:,:,n) = sum(flx,dim=3)/real(nread)
    deallocate( flx )

    !--- ocean fluxes

    ! number of fields to read for each flux time step
    nread = int(statres*24./real(nt_ocn))
    if ( nread.eq.0 ) nread = 1
    print*, 'init_co2: ocn nread = ',nread
    allocate( flx(nxgrid,nygrid,nread), stat = ierr )
    if ( ierr.ne.0 ) stop 'ERROR init_co2: not enough memory'
    filename = str_replace(files%filename_ocn, 'YYYY', adate)
    if ( index(filename, 'MM').ne.0 ) then
      filename = str_replace(filename, 'MM', amonth)
    endif
    filename = trim(files%path_prior)//trim(filename)
    inquire(file=trim(filename),exist=lexist)
    if (.not.lexist) then    
      write(logid,*) 'ERROR: cannot find '//trim(filename)
      stop
    endif
    write(logid,*) 'Reading ocean :'//trim(filename)
    call read_ncdf(filename, &
                   files%varname_ocn, &
                   files%lonname_ocn, &
                   files%latname_ocn, &
                   files%timename_ocn,&
                   llx, lly, &
                   nxgrid, nygrid, &
                   jd, nread, num, &
                   flx)
    print*, 'init_co2: ocn num = ',num
    ! convert to kg/m2/s
    flx = flx/3600.
    ! apply numerical scaling
    flx = flx*numscale
    ! average field to flux time step 
    fluxes%flxocn(:,:,n) = sum(flx,dim=3)/real(nread)
    deallocate( flx )

    ! flux time stamp is start of time step
    fluxes%time(n) = jd

    jd = jd + dble(statres)

  end do 
  print*, 'init_co2: fluxes%time = ',fluxes%time

  ! regional prior fluxes
  ! ---------------------

  ! these fluxes are used to calculate the contribution from fluxes inside the domain 
  ! fossil fuel and NEE fluxes are averaged/interpolated to temporal resolution of the footprints 
  ! ocean fluxes are average/interpolated to temporal resolution of state vector

  ! note: if nested is false use same fluxes as for global but 
  ! average/interpolate to temporal resolution of the footprints

  if ( .not.config%nested ) then
    lenstr=len_trim(files%filenest_nee)
    if ( lenstr.eq.0 ) then
      files%filenest_nee = files%filename_nee
      files%varnest_nee = files%varname_nee
      files%lonnest_nee = files%lonname_nee
      files%latnest_nee = files%latname_nee
      files%timenest_nee = files%timename_nee
    endif
    lenstr=len_trim(files%filenest_ff)
    if (lenstr.eq.0) then
      files%filenest_ff = files%filename_ff
      files%varnest_ff = files%varname_ff
      files%lonnest_ff = files%lonname_ff
      files%latnest_ff = files%latname_ff
      files%timenest_ff = files%timename_ff
    endif
  endif

  !--- fossil fuel

  print*, 'nday = ',nday
  print*, 'nt_ff_reg = ',nt_ff_reg
  fluxes%flxff_nest(:,:,:) = 0.
  print*, 'init_co2: ffreg timesteps input to read = ',(nday*24)/nt_ff_reg
  allocate( flxff(nxregrid,nyregrid,(nday*24)/nt_ff_reg), stat = ierr )
  if ( ierr.ne.0 ) stop 'ERROR init_co2: not enough memory'
  allocate( timeff((nday*24)/nt_ff_reg) )
  jd = juldatei
  n = 0
  ! loop over months reading all input fluxes
  do while ( jd.le.juldatef )
    call caldate(jd, jjjjmmdd, hhmiss)
    write(adate,'(I4.4)') jjjjmmdd/10000
    write(amonth,'(I2.2)') (jjjjmmdd-(jjjjmmdd/10000)*10000)/100
    eomday = calceomday(jjjjmmdd/100)
    eomday_save = eomday
    ! no leap years in input data
    if ( eomday.eq.29 ) eomday = 28
    print*, 'init_co2: ffreg eomday = ',eomday
    ! number of fields to read this month
    nread = eomday*24/nt_ff_reg
    print*, 'init_co2: ffreg nread = ',nread
    if ( nread.eq.0 ) nread = 1
    allocate( flx(nxregrid,nyregrid,nread), stat = ierr )
    if ( ierr.ne.0 ) stop 'ERROR init_co2: not enough memory'
    filename = str_replace(files%filenest_ff, 'YYYY', adate)
    if ( index(filename, 'MM').ne.0 ) then
      filename = str_replace(filename, 'MM', amonth)
    endif
    filename = trim(files%path_prior)//trim(filename)
    inquire(file=trim(filename),exist=lexist)
    if (.not.lexist) then
      write(logid,*) 'ERROR: cannot find '//trim(filename)
      stop
    endif
    write(logid,*) 'Reading fossil fuel :'//trim(filename)
    if ( nt_ff_reg.lt.24 ) then
      ! assume timestamp is hours since 1-Jan each year
      jdi = (jd - juldate((jjjjmmdd/10000)*10000+101,0))*24d0 + juldate(19000101,0)
    else
      jdi = jd
    endif
    call read_ncdf(filename, &
                   files%varnest_ff, &
                   files%lonnest_ff, &
                   files%latnest_ff, &
                   files%timenest_ff,&
                   rllx, rlly, &
                   nxregrid, nyregrid, &
                   jdi, nread, num, &
                   flx)
    print*, 'init_co2: ffreg num = ',num
    ! timestamp julian days
    do i = 1, nread 
      if ( (n+i).gt.int(real(nday*24)/real(nt_ff_reg)) ) exit
      timeff(n+i) = jd + dble((i-1)*nt_ff_reg)/24d0
      flxff(:,:,n+i) = flx(:,:,i)
    end do
    deallocate( flx )
    n = n + nread
    jd = jd + dble(eomday_save)
  end do
  n = min(n, int(real(nday*24)/real(nt_ff_reg)))
  print*, 'init_co2: ffreg n = ',n
  print*, 'init_co2: timeff = ',timeff
 
  ! convert from input flux units to kg/m2/s
  flxff = flxff*config%coeff_ff_reg/3600.
  ! apply numerical scaling
  flxff = flxff*numscale
 
  ! average/interpolate to footprint timestamp
  allocate( datain(nxregrid,n) )
  allocate( dataout(nxregrid,nday*ndgrid) )
  allocate( timefp(nday*ndgrid) )
  do i = 1, nday*ndgrid
    timefp(i) = juldatei + dble(i-1)/dble(ndgrid)
  enddo
  print*, 'init_co2: timefp = ',timefp
  if ( ndgrid.lt.(24/nt_ff_reg) ) then
    ! average flux
    do jy = 1, nyregrid
      datain = flxff(:,jy,:)
      call average(nxregrid, n, timeff, datain, nday*ndgrid, timefp, dataout)
      fluxes%flxff_nest(:,jy,:) = dataout(:,:)
    end do
  else if ( ndgrid.gt.(24/nt_ff_reg) ) then
    ! interpolate flux
    do jy = 1, nyregrid
      datain = flxff(:,jy,:)
      call interp(nxregrid, n, timeff, datain, nday*ndgrid, timefp, dataout)
      fluxes%flxff_nest(:,jy,:) = dataout(:,:)
    end do
  else if ( ndgrid.eq.(24/nt_ff_reg) ) then
    fluxes%flxff_nest(:,:,:) = flxff
  endif
  fluxes%timefp(:) = timefp
  deallocate( datain )
  deallocate( dataout )
  deallocate( timefp )
  deallocate( flxff )
  deallocate( timeff )

  !--- ocean fluxes

  ! ocean contribution from ocean fluxes inside the inversion domain
  ! is calculated at same resolution as state vector and the background
  ! contribution, hence, only need to do following if nested = true

  if ( config%nested ) then
    write(logid,*) 'Reading regional ocean fluxes'
    ! nested
    fluxes%flxocn_nest(:,:,:) = 0.
    numocn = max(1,(nday*24)/nt_ocn)
    print*, 'init_co2: ocn reg timesteps input to read = ',numocn
    allocate( flxocn(nxregrid,nyregrid,numocn), stat = ierr )
    if ( ierr.ne.0 ) stop 'ERROR init_co2: not enough memory'
    allocate( timeocn(numocn) )
    n = 0
    jd = juldatei
    do while ( jd.le.juldatef )
      call caldate(jd, jjjjmmdd, hhmiss)
      write(adate,'(I4.4)') jjjjmmdd/10000
      eomday = calceomday(jjjjmmdd/100)
      eomday_save = eomday
      ! no leap years in input data
      if ( eomday.eq.29 ) eomday = 28
      ! number of fields to read this month
      nread = eomday*24/nt_ocn
      if ( nread.eq.0 ) nread = 1
      allocate( flx(nxregrid,nyregrid,nread), stat = ierr )
      if ( ierr.ne.0 ) stop 'ERROR init_co2: not enough memory'
      filename = str_replace(files%filenest_ocn, 'YYYY', adate)
      if ( index(filename, 'MM').ne.0 ) then
        filename = str_replace(filename, 'MM', amonth)
      endif
      filename = trim(files%path_prior)//trim(filename)
      inquire(file=trim(filename),exist=lexist)
      if (.not.lexist) then
        write(logid,*) 'ERROR: cannot find '//trim(filename)
        stop
      endif
      write(logid,*) 'Reading regional ocean :'//trim(filename)
      call read_ncdf(filename, &
                     files%varnest_ocn, &
                     files%lonnest_ocn, &
                     files%latnest_ocn, &
                     files%timenest_ocn,&
                     rllx, rlly, &
                     nxregrid, nyregrid, &
                     jd, nread, num, &
                     flx)
      ! timestamp in julian days
      do i = 1, nread
        if ( (n+i).gt.numocn ) exit
        timeocn(n+i) = jd + dble((i-1)*nt_ocn)/24d0
        flxocn(:,:,n+i) = flx(:,:,i)
      end do
      deallocate( flx )
      n = n + nread
      jd = jd + dble(eomday_save)
    end do
    n = min(n, numocn)
    print*, 'init_co2: ocnreg n = ',n
    print*, 'init_co2: timeocn = ',timeocn

    ! convert to kg/m2/s
    flxocn = flxocn/3600.
    ! apply numerical scaling
    flxocn = flxocn*numscale

    ! average/interpolate to state vector timestamp
    allocate( datain(nxregrid,n) )
    allocate( dataout(nxregrid,ntstate) )
    if ( statres.gt.real(nt_ocn/24) ) then
      ! average flux
      do jy = 1, nyregrid
        datain = flxocn(:,jy,:)
        call average(nxregrid, n, timeocn, datain, ntstate, fluxes%time, dataout)
        fluxes%flxocn_nest(:,jy,:) = dataout(:,:)
      end do
    else if ( statres.lt.real(nt_ocn/24) ) then
      ! interpolate flux
      do jy = 1, nyregrid
        datain = flxocn(:,jy,:)
        call interp(nxregrid, n, timeocn, datain, ntstate, fluxes%time, dataout)
        fluxes%flxocn_nest(:,jy,:) = dataout(:,:)
      end do
    else if ( statres.eq.real(nt_ocn/24) ) then
      fluxes%flxocn_nest(:,:,:) = flxocn(:,:,:)
    endif
    deallocate( datain )
    deallocate( dataout )
    deallocate( flxocn )
    deallocate( timeocn )

  endif

  !--- NEE fluxes 

  ! read regional NEE and average/interpolate to footprint timestamp

  write(logid,*) 'Reading hourly NEE'

  fluxes%flxnee_nest(:,:,:) = 0.

  allocate( flxnee(nxregrid,nyregrid,nday*nd_nee), stat = ierr )
  if ( ierr.ne.0 ) stop 'ERROR init_co2: not enough memory'
  allocate( timenee(nday*nd_nee) )
  n = 0
  jd = juldatei
  do while ( jd.le.juldatef )
    call caldate(jd, jjjjmmdd, hhmiss)
    write(adate,'(I4.4)') jjjjmmdd/10000
    eomday = calceomday(jjjjmmdd/100)
    eomday_save = eomday
    print*, 'init_co2: eomday = ',eomday
    ! no leap years in input data
    if ( eomday.eq.29 ) eomday = 28
    ! number of fields to read this month
    nread = eomday*nd_nee
    if ( nread.eq.0 ) nread = 1
    print*, 'init_co2: neereg nread = ',nread
    allocate( flx(nxregrid,nyregrid,nread), stat = ierr )
    filename = str_replace(files%filenest_nee, 'YYYY', adate)
    if ( index(filename, 'MM').ne.0 ) then
      filename = str_replace(filename, 'MM', amonth)
    endif
    filename = trim(files%path_prior)//trim(filename)
    inquire(file=trim(filename),exist=lexist)
    if (.not.lexist) then
      write(logid,*) 'ERROR: cannot find '//trim(filename)
      stop
    endif
    call read_ncdf(filename, &
                   files%varnest_nee, &
                   files%lonnest_nee, &
                   files%latnest_nee, &
                   files%timenest_nee,&
                   rllx, rlly, &
                   nxregrid, nyregrid, &
                   jd, nread, num, &
                   flx)
    do i = 1, nread
      if ( (n+i).gt.(nday*nd_nee) ) exit
      timenee(n+i) = jd + dble((i-1)*nt_nee)/24d0
      flxnee(:,:,n+i) = flx(:,:,i)
    end do
    deallocate( flx )
    n = n + nread
    print*, 'init_co2: n = ',n
    jd = jd + dble(eomday_save)
  end do
  n = min(n, (nday*nd_nee))
  print*, 'init_co2: neereg n = ',n
  print*, 'init_co2: timenee = ',timenee
  print*, 'init_co2: ndgrid, nd_nee = ',ndgrid, nd_nee

  ! convert to kg/m2/s
  flxnee = flxnee/3600.
  ! apply numerical scaling
  flxnee = flxnee*numscale
    
  ! average/interpolate to footprint time
  allocate( datain(nxregrid,n) )
  allocate( dataout(nxregrid,nday*ndgrid) )
  if ( ndgrid.lt.nd_nee ) then
    ! average flux
    print*, 'init_co2: averaging nee to footprint time resolution'
    do jy = 1, nyregrid
      datain = flxnee(:,jy,:)
      call average(nxregrid, n, timenee, datain, nday*ndgrid, fluxes%timefp, dataout)
      fluxes%flxnee_nest(:,jy,:) = dataout(:,:)
    end do
  else if ( ndgrid.gt.nd_nee ) then
    ! interpolate flux
    print*, 'init_co2: interpolating nee to footprint time resolution'
    do jy = 1, nyregrid
      datain = flxnee(:,jy,:)
      call interp(nxregrid, n, timenee, datain, nday*ndgrid, fluxes%timefp, dataout)
      fluxes%flxnee_nest(:,jy,:) = dataout(:,:)
    end do
  else if ( ndgrid.eq.nd_nee ) then
    print*, 'init_co2: fluxes have already footprint time resolution'
    fluxes%flxnee_nest(:,:,:) = flxnee(:,:,:)
  endif
  deallocate( datain )
  deallocate( dataout )
  deallocate( timenee )

  ! uncertainty for NEE
  ! -------------------

  ! calculate state vector uncertainty for NEE
  ! on aggregated grid and for statres and statres_hr timestamps

  err_box(:,:,:) = 0.
  nee_box(:,:,:) = 0.

  do nts = 1, ntstate

    ! average hourly NEE over statres dimension
    reg_nee(:,:,:) = 0.
    do i = 1, int(statres)
      n = (nts-1)*int(statres) + i
      reg_nee(:,:,:) = reg_nee(:,:,:) + flxnee(:,:,(n-1)*nd_nee+1:n*nd_nee)
    end do
    reg_nee(:,:,:) = reg_nee(:,:,:)/statres

    ! calculate area-weighted mean NEE for each box
    do n = 1, nd_nee
      do jy = 1, nyregrid
        area = grid_area(reg_lat(jy)+0.5*rdy, rdy, rdx)
        do ix = 1, nxregrid
          if ( nbox_xy(ix,jy).gt.0 ) then
            nee_box(nbox_xy(ix,jy),nts,n) = nee_box(nbox_xy(ix,jy),nts,n) + reg_nee(ix,jy,n) * area
          endif
        end do
      end do
      nee_box(:,nts,n) = nee_box(:,nts,n)/area_box
    end do

    ! calculate prior NEE uncertainty for each box
    do nb = 1, nbox
      cnt(:) = 0
      do n = 1, nd_nee
        hh = (n-1)*24/nd_nee + int(hloc(nb))
        if ( hh.lt.0 ) then
          hh = hh + 24
        else if ( hh.ge.24 ) then
          hh = hh - 24
        endif
        do nt = 1, ndt
          if ( (hh.ge.(nt-1)*24/ndt).and.(hh.lt.nt*24/ndt) ) then
            cnt(nt) = cnt(nt) + 1
            err_box(nb,nts,nt) = err_box(nb,nts,nt) + nee_box(nb,nts,n)*config%flxerr
          endif
        end do
      end do
      do nt = 1, ndt
        err_box(nb,nts,nt) = err_box(nb,nts,nt)/cnt(nt)
      end do
    end do

  end do

  ! verbose output for testing
  ! --------------------------

  if ( config%verbose ) then
    ! NEE for region interpolated to footprint times
    filename = trim(files%path_output)//'flxnee_nest.txt'
    write(rowfmt,'(A,I6,A)') '(',nxregrid,'(E11.4,1X))'
    open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
    do nd = 1, nday*ndgrid
      do n = 1, nyregrid
        write(100,fmt=rowfmt) fluxes%flxnee_nest(:,n,nd)
      end do
    end do
    close(100)
    ! NEE for region before interpolation (first statres timestep only)
    filename = trim(files%path_output)//'flxnee.txt'
    write(rowfmt,'(A,I6,A)') '(',nxregrid,'(E11.4,1X))'
    open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
    do nd = 1, int(statres)*nd_nee
      do n = 1, nyregrid
        write(100,fmt=rowfmt) flxnee(:,n,nd)
      end do
    end do
    close(100)  
    ! NEE for boxes
    filename = trim(files%path_output)//'nee_box.txt'
    write(rowfmt,'(A,I6,A)') '(',nbox,'(E11.4,1X))'
    open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
    do nts = 1, ntstate
      do n = 1, nd_nee
        write(100,fmt=rowfmt) nee_box(:,nts,n)
      end do
    end do
    close(100)
    ! NEE uncertainty for boxes  
    write(rowfmt,'(A,I6,A)') '(',nbox,'(E11.4,1X))'
    open(100,file=trim(files%path_output)//'err_box.txt',status='replace',action='write',iostat=ierr)
    do nts = 1, ntstate
      do n = 1, ndt
        write(100,fmt=rowfmt) err_box(:,nts,n)
      end do
    end do
    close(100)
  endif
 
  deallocate( flxnee )

  ! prior state vector
  ! ------------------

  ! number of state variables
  npvar = ndt*ntstate*nbox
  ndvar = ndt*nbox
  if ( config%opt_cini ) then
    nvar = npvar + ntcini*ncini
  else
    nvar = npvar
  endif

  write(logid,*) 'Number of flux state variables: ',npvar
  write(logid,*) 'Number of state variables per flux time step: ',ndvar
  write(logid,*) 'Total number of state variables: ',nvar

  call alloc_states(states) 
  allocate( xerr(ndvar) )

  ! assign prior state vector
  states%px0(:) = 0.
  if ( config%opt_cini ) states%px0(npvar+1:nvar) = 1.

  ! assign time stamps to flux time steps
  states%xtime(:) = fluxes%time(:)
  print*, 'init_co2: states%xtime = ',states%xtime

  ! assign time stamps to mixing ratio scalar time steps
  do n = 1, ntcini
    states%cinitime(n) = juldatei + dble((n-1)*cinires) - trajdays
  end do
  print*, 'init_co2: cinitime = ',states%cinitime

  ! assign best guess of state vector
  if ( (config%prior_bg.eq.1).and.(config%method.eq.'congrad') ) then
    ! from previous inversion
    call read_bg(files, config, states)
  else
    ! from prior
    states%px(:) = states%px0(:)
  endif

  ! prior state vector uncertainty
  do n = 1,ndt
    xerr((n-1)*nbox+1:n*nbox) = abs(sum(err_box(:,:,n),dim=2)/ntstate)
  end do

  if ( config%verbose ) then
    ! state vector uncertainty   
    open(100,file=trim(files%path_output)//'xerr.txt',status='replace',action='write',iostat=ierr)
    write(rowfmt,'(A,I6,A)') '(',nbox,'(E11.4,1X))'
    do n = 1, ndt
      write(100,fmt=rowfmt) xerr((n-1)*nbox+1:n*nbox)
    end do
    close(100)
  endif

  ! correlation and error covariance matrices
  ! -----------------------------------------

  corr(:,:) = 0.
  if ( config%spa_corr ) then
    write(logid,*) 'Using spatial correlation in B'
    ! check if file already exists (in case of rerun)
    inquire(file=trim(files%path_output)//'correl.txt',exist=lexist)
    if ( lexist ) then
      write(logid,*) 'Reading spatial correlation from correl.txt'
      open(100,file=trim(files%path_output)//'correl.txt',action='read',status='old',iostat=ierr)
      write(rowfmt,'(A,I6,A)') '(',nbox,'(E11.4,1X))'
      do nb = 1, nbox
        read(100,rowfmt) corr(nb,:)
      end do
      close(100)
    else
      call correl(files, config, corr)
    endif
  else
    do n = 1, nbox
      corr(n,n) = 1.
    end do
  endif
  call error_cov(config, files, states, covar, corr, xerr)

  ! observations
  ! ------------

  nobs = 0
  call read_obs(config, files)
  call assign_obs(config, files, obs)

  ! initial concentrations
  ! ----------------------

  ! initial concentrations from termination of trajectories
  ! only do for a fresh run
  if ( config%restart.eq.0 ) then
    if ( files%init_type.lt.5 ) then
      call init_cini(files, config, obs)
    else
      call init_cini_month(files, config, obs)
    endif
  endif
    
  ! add random perturbation for MC
  ! ------------------------------
   
  if ( config%run_mode.eq.2 ) then
    write(logid,*) 'Adding random perturbation to state vector and observations...'
    ! perturb state vector
    call perturb_state(config, states, covar)
    ! perturb observations 
    call perturb_obs(config, obs)
  endif

  ! save datatype obs for restart
  ! -----------------------------

  if ( config%restart.eq.1 ) then
    ! read datatype obs from file
    write(logid,*) 'WARNING init_co2: reading cini from previous run'
    open(100,file=trim(files%path_output)//'obsdatatype.txt',status='old',action='read',iostat=ierr)
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR init_co2: cannot open file obsdatatype.txt'
      stop
    endif
    allocate( recs(nobs) )
    allocate( jdate(nobs) )
    write(rowfmt,'(A,I1,A,I6,A)') '(A',recname_len,',1X,F14.6,1X,F10.4,1X,',ncini,'(F10.4,1X))'
    read(100,*) dump
    do n = 1, nobs
      read(100,rowfmt,iostat=ierr) recs(n), jdate(n), conc, obs%cini(n,:)
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR init_co2: reading obsdatatype.txt'
        exit
      endif
    end do
    ! check consistency
    if ( recs(1).ne.obs%recs(1).or.recs(nobs).ne.obs%recs(nobs).or.jdate(1).ne.obs%jdate(1).or.jdate(nobs).ne.obs%jdate(nobs) ) then
      write(logid,*) 'ERROR init_co2: data in obsdatatype.txt inconsistent with run'
      print*, recs(1), obs%recs(1)
      print*, recs(nobs), obs%recs(nobs)
      print*, jdate(1), obs%jdate(1)
      print*, jdate(nobs), obs%jdate(nobs)
      stop
    endif  
    deallocate( recs )
    deallocate( jdate )
  else
    ! save datatype obs
    call save_for_restart(files, obs)  
  endif

end subroutine init_co2


