!---------------------------------------------------------------------------------------
! FLEXINVERT: mod_tools
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_tools
!!
!! Purpose:    Module containing numerical functions.
!! 
!---------------------------------------------------------------------------------------


module mod_tools

  use mod_var

  implicit none

  private

  public :: sort, unique, grid_area, interp, average

  contains

  ! --------------------------------------------------
  ! sort
  ! --------------------------------------------------
  !> sort
  !!
  !! Purpose:  Sort the elements in a vector in 
  !!           ascending order.
  !! 
  ! --------------------------------------------------

  subroutine sort(n, arr, ind)

    implicit none

    integer,                    intent(in)     :: n
    real(kind=8), dimension(n), intent(in out) :: arr
    integer, dimension(n),      intent(in out) :: ind

    integer, parameter                         :: m=7, nstack=50
    integer                                    :: i, ir, j, k, l, jstack
    integer, dimension(nstack)                 :: istack
    real(kind=8)                               :: a, temp
    integer                                    :: ia, itemp

      do i=1,n
        ind(i)=i
      end do
      jstack=0
      l=1
      ir=n
1     if(ir-l.lt.m) then
        do j=l+1,ir
          a=arr(j)
          ia=ind(j)
          do i=j-1,1,-1
            if(arr(i).le.a) go to 2
            arr(i+1)=arr(i)
            ind(i+1)=ind(i)
          end do
          i=0
2         arr(i+1)=a
          ind(i+1)=ia
        end do
        if(jstack.eq.0) return
        ir=istack(jstack)
        l=istack(jstack-1)
        jstack=jstack-2
      else
        k=(l+ir)/2
        temp=arr(k)
        itemp=ind(k)
        arr(k)=arr(l+1)
        arr(l+1)=temp
        ind(k)=ind(l+1)
        ind(l+1)=itemp
        if(arr(l+1).gt.arr(ir)) then
          temp=arr(l+1)
          itemp=ind(l+1)
          arr(l+1)=arr(ir)
          arr(ir)=temp
          ind(l+1)=ind(ir)
          ind(ir)=itemp
        endif
        if(arr(l).gt.arr(ir)) then
          temp=arr(l)
          itemp=ind(l)
          arr(l)=arr(ir)
          arr(ir)=temp
          ind(l)=ind(ir)
          ind(ir)=itemp
        endif
        if(arr(l+1).gt.arr(l)) then
          temp=arr(l+1)
          itemp=ind(l+1)
          arr(l+1)=arr(l)
          arr(l)=temp
          ind(l+1)=ind(l)
          ind(l)=itemp
        endif
        i=l+1
        j=ir
        a=arr(l)
        ia=ind(l)
3       continue
          i=i+1
        if(arr(i).lt.a) go to 3
4       continue
          j=j-1
        if(arr(j).gt.a) go to 4
        if(j.lt.i) go to 5
        temp=arr(i)
        itemp=ind(i)
        arr(i)=arr(j)
        arr(j)=temp
        ind(i)=ind(j)
        ind(j)=itemp
        go to 3
5       arr(l)=arr(j)
        arr(j)=a
        ind(l)=ind(j)
        ind(j)=ia
        jstack=jstack+2
        if(jstack.gt.nstack) print*, 'ERROR sort: nstack too small'
        if(ir-i+1.ge.j-l) then
          istack(jstack)=ir
          istack(jstack-1)=i
          ir=j-1
        else
          istack(jstack)=j-1
          istack(jstack-1)=l
          l=i
        endif
      endif
      goto 1

  end subroutine sort

  ! --------------------------------------------------
  ! unique
  ! --------------------------------------------------
  !> unique
  !!
  !! Purpose:  Returns the unique elements of a 
  !!           vector.
  !!
  ! --------------------------------------------------

  subroutine unique(num, arr, stack, cnt)

    implicit none

    integer,              intent(in)   :: num
    real, dimension(num), intent(in)   :: arr
    real, dimension(num), intent(out)  :: stack
    integer,              intent(out)  :: cnt

    integer                            :: i, k

    stack(:) = 0.
    k = 0
    do i = 1, num
      if ( .not.(any(stack.eq.arr(i))) ) then
        k = k + 1
        stack(k) = arr(i)
      endif
    enddo
    cnt = k

  end subroutine unique

  ! --------------------------------------------------
  ! grid_area
  ! --------------------------------------------------
  !> grid_area
  !!
  !! Purpose:  Calculates the area of a grid cell.
  !!
  ! --------------------------------------------------

  function grid_area(ylat, dy, dx)

    implicit none

    real, intent(in) :: ylat, dy, dx
    real             :: grid_area

    real, parameter  :: pi=3.14159265, rearth=6.371e6, pih=pi/180.
    real             :: ylata, ylatp, ylatm, cosfact, cosfactp, cosfactm, hzone

    ylata = ylat
    ylatp = ylat+0.5*dy
    ylatm = ylat-0.5*dy
    if (ylatm.lt.0.and.ylatp.gt.0) then
      hzone = rearth*pih
    else
      cosfact = COS(ylata*pih)*rearth
      cosfactp = COS(ylatp*pih)*rearth
      cosfactm = COS(ylatm*pih)*rearth
      if(cosfactp.lt.cosfactm) then
        hzone = SQRT(rearth**2-cosfactp**2)-SQRT(rearth**2-cosfactm**2)
      else
        hzone = SQRT(rearth**2-cosfactm**2)-SQRT(rearth**2-cosfactp**2)
      endif
    endif
    grid_area = 2.*pi*rearth*hzone*dx/360.

  end function grid_area

  ! --------------------------------------------------
  ! interp
  ! --------------------------------------------------
  !> interp
  !!
  !! Purpose: Interpolate a data vector from an input
  !!          timestamp to an output timestamp
  !!
  ! --------------------------------------------------

  subroutine interp(nx, nin, timein, datain, nout, timeout, dataout)

    implicit none

    integer,                       intent(in)  :: nx
    integer,                       intent(in)  :: nin
    real(kind=8), dimension(nin),  intent(in)  :: timein
    real, dimension(nx,nin),       intent(in)  :: datain
    integer,                       intent(in)  :: nout
    real(kind=8), dimension(nout), intent(in)  :: timeout
    real, dimension(nx,nout),   intent(in out) :: dataout

    integer                                    :: n
    integer                                    :: ii, i1, i2

    ! interpolate data vector on intput timestamp, timein,
    ! to output timestamp, timeout
    dataout(:,:) = 0.
    do n = 1, nout
      ii = minloc(timein,dim=1,mask=abs(timein-timeout(n)).eq.minval(abs(timein-timeout(n))))
      if ( timein(ii).gt.timeout(n) ) then
        i1 = max(1, ii-1)
        i2 = ii
      else
        i1 = ii
        i2 = min(nin, (ii+1))
      endif
      if ( i1.ne.i2 ) then
        dataout(:,n) = datain(:,i1) + &
                        (datain(:,i2) - datain(:,i1)) * sngl((timeout(n) - timein(i1)) / &
                        (timein(i2) - timein(i1)))
      else
        dataout(:,n) = datain(:,i1)
      endif
    end do

   end subroutine interp

  ! --------------------------------------------------
  ! average
  ! --------------------------------------------------
  !> average
  !!
  !! Purpose: Average a data vector on an input 
  !!          timestamp to an output timestamp
  !!
  ! --------------------------------------------------

  subroutine average(nx, nin, timein, datain, nout, timeout, dataout)

    implicit none

    integer,                       intent(in)  :: nx
    integer,                       intent(in)  :: nin
    real(kind=8), dimension(nin),  intent(in)  :: timein
    real, dimension(nx,nin),       intent(in)  :: datain
    integer,                       intent(in)  :: nout
    real(kind=8), dimension(nout), intent(in)  :: timeout
    real, dimension(nx,nout),      intent(out) :: dataout

    integer                                    :: n
    real                                       :: fin, fout
    integer                                    :: i0, i1, i2
    real                                       :: frac, rem1, rem2

    ! check timeseries frequency
    fin = 1./sngl(timein(2) - timein(1))
    fout = 1./sngl(timeout(2) - timeout(1))
    if ( fin.lt.fout ) then
      print*, 'ERROR average: input timeseries less frequent than output'
      stop
    endif

    ! average data vector on input timestamp, timein, 
    ! to output timestamp, timeout
    frac = fin/fout
    i0 = minloc(timein,dim=1,mask=abs(timein-timeout(1)).eq.minval(abs(timein-timeout(1))))
    do n = 1, nout
      i1 = i0 + int((real(n)-1.)*frac)
      i2 = i0 + int(real(n)*frac) - 1
      rem1 = (real(n)-1.)*frac - real(int((real(n)-1.)*frac))
      rem2 = real(n)*frac - real(int(real(n)*frac))
      i1 = max(1, i1)
      i2 = min(nin, i2)
      if ( rem1.gt.0. ) then
        if ( i2.lt.nin ) then
          dataout(:,n) = (sum(datain(:,(i1+1):i2),dim=2) + rem1*datain(:,i1) + rem2*datain(:,i2+1)) / &
                           (i2 - i1 + rem1 + rem2)
        else
          dataout(:,n) = (sum(datain(:,(i1+1):i2),dim=2) + rem1*datain(:,i1)) / &
                           (i2 - i1 + rem1)
        endif
      else
        if ( i2.lt.nin ) then
          dataout(:,n) = (sum(datain(:,i1:i2),dim=2) + rem2*datain(:,i2+1)) / &
                           (i2 - i1 + 1 + rem2)
        else
          dataout(:,n) = sum(datain(:,i1:i2),dim=2) / (i2 - i1 + 1)
        endif
      endif
    end do

  end subroutine average

  ! --------------------------------------------------

end module mod_tools


