!---------------------------------------------------------------------------------------
! FLEXINVERT: read_obs
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> read_obs
!!
!! Purpose:    Read observations from files and write to obsread.txt
!!
!---------------------------------------------------------------------------------------

subroutine read_obs(config, files)

  use netcdf
  use mod_settings
  use mod_var
  use mod_dates
  use mod_strings
  use mod_save,    only : check

  implicit none

  type (config_t), intent(in)                            :: config
  type (files_t),  intent(in)                            :: files
  character(len=max_path_len), dimension(:), allocatable :: filelist
  character(len=max_path_len)                            :: rowfmt, numfmt, filerelease
  character(len=max_path_len), dimension(4)              :: args
  integer                                                :: narg
  character(len=recname_len)                             :: recs
  character(len=50)                                      :: species, before, sep
  character(len=200)                                     :: header, string
  character(len=8)                                       :: adate
  integer                                                :: ierr
  integer                                                :: cnt
  integer                                                :: i, nr, nf, nfiles, strlen
  integer                                                :: yyyymmdd, hhmiss, hl, hh
  real(kind=8)                                           :: jdate, avetime
  real                                                   :: conc, err, measerr, lon, alt
  integer                                                :: num, reclen, nretr, nrelease, nlayer
  integer                                                :: ncid, varid, dimid
  integer, dimension(:), allocatable                     :: idate, itime
  real, dimension(:), allocatable                        :: cpri, cakpri, cdryair, vmr, vmr_err
  real                                                   :: xcpri, xcakpri, scl_cpri, scl_cakpri, scl_cdryair !, cak_av 
  logical                                                :: lexist

  ! list observation files
  ! ----------------------

  print*, 'readobs: files%path_output = ',trim(files%path_output)
  print*, 'readobs: files%path_satobs = ',trim(files%path_satobs)

  if ( config%ground.and.config%satellite ) then
    call system('ls '//trim(files%path_obs)//'*'//trim(files%suffix)//&
                ' '//trim(files%path_satobs)//'*.nc'//' | wc -l > '//trim(files%path_output)//'obsfiles.txt')
    call system('ls '//trim(files%path_obs)//' | grep '//trim(files%suffix)//' >> '//trim(files%path_output)//'obsfiles.txt')
    call system('ls '//trim(files%path_satobs)//' | grep .nc >> '//trim(files%path_output)//'obsfiles.txt')
  else if ( config%ground ) then
    call system('ls '//trim(files%path_obs)//'*'//trim(files%suffix)//' | wc -l > '//trim(files%path_output)//'obsfiles.txt')
    call system('ls '//trim(files%path_obs)//' | grep '//trim(files%suffix)//' >> '//trim(files%path_output)//'obsfiles.txt')
  else if ( config%satellite ) then
    call system('ls '//trim(files%path_satobs)//'*.nc'//' | wc -l > '//trim(files%path_output)//'obsfiles.txt')
    call system('ls '//trim(files%path_satobs)//' | grep .nc >> '//trim(files%path_output)//'obsfiles.txt')
  endif

  open(100,file=trim(files%path_output)//'obsfiles.txt',action='read',status='old',iostat=ierr)
  if ( ierr.ne. 0 ) then
    write(logid,*) 'ERROR: cannot open obsfiles.txt'
    stop
  endif
  read(100,*,iostat=ierr) nfiles
  allocate ( filelist(nfiles), stat = ierr )
  do nf = 1, nfiles
    read(100,fmt='(A)',iostat=ierr) filelist(nf)
    if (ierr.ne.0) exit
  end do
  close(100)

  ! loop over files
  ! ---------------

  ! output file for all observations
  open(200,file=trim(files%path_output)//'obsread.txt',action='write',status='old',iostat=ierr)
  if (ierr.ne.0) then
    write(logid,*) 'ERROR: cannot open obsread.txt'
    stop
  endif
  write(200,*) 'rec yyyymmdd hhmmss juldate avetime conc error cpri cakpri cdryair'

  cnt = 0
  do nf = 1, nfiles

    ! check if ground-based or satellite observation
    string = filelist(nf)
    print*, 'readobs: string = ',string
    call split(string,".",before,sep)
    print*, 'readobs: before, string = ',before, string

    if ( string.eq.'nc' ) then

      ! satellite   
      ! ---------

      ! check if data is in inversion time interval
      call parse_string(before,"_",args,narg)
      adate = args(narg)
      print*, 'read_obs: adate = ',adate
      read(adate,*) yyyymmdd
      jdate = juldate(yyyymmdd, 0)
      if ( jdate.lt.juldatei.or.jdate.gt.juldatef ) cycle

      ! read ncdf file
      call check( nf90_open(trim(files%path_satobs)//trim(filelist(nf)),nf90_NOWRITE,ncid) )
      call check( nf90_inq_dimid(ncid,'retrieval',dimid) )
      call check( nf90_inquire_dimension(ncid,dimid,len=nretr) )
      allocate( idate(nretr), stat=ierr )
      allocate( itime(nretr), stat=ierr )
      allocate( cpri(nretr), stat=ierr )
      allocate( cakpri(nretr), stat=ierr )
      allocate( cdryair(nretr), stat=ierr )
      allocate( vmr(nretr), stat=ierr )
      allocate( vmr_err(nretr), stat=ierr )
      call check( nf90_inq_varid(ncid,'idate',varid) )
      call check( nf90_get_var(ncid,varid,idate) )
      call check( nf90_inq_varid(ncid,'itime',varid) )
      call check( nf90_get_var(ncid,varid,itime) )
      call check( nf90_inq_varid(ncid,'cpri',varid) )
      call check( nf90_get_var(ncid,varid,cpri) )
      call check( nf90_inq_varid(ncid,'cakpri',varid) )
      call check( nf90_get_var(ncid,varid,cakpri) )
      call check( nf90_inq_varid(ncid,'cdryair',varid) )
      call check( nf90_get_var(ncid,varid,cdryair) )
      call check( nf90_inq_varid(ncid,'vmr',varid) )
      call check( nf90_get_var(ncid,varid,vmr) )
      call check( nf90_inq_varid(ncid,'vmr_err',varid) )
      call check( nf90_get_var(ncid,varid,vmr_err) )
      call check( nf90_close(ncid) )

      ! loop over retrievals
      avetime = 0d0
      write(numfmt,fmt='(A,I1,A,I1,A)') '(I',recname_len,'.',recname_len,')'
      write(rowfmt,fmt='(A,I1,A)') '(A',recname_len,',1X,I8,1X,I6,1X,F14.6,1X,F10.6,1X,F10.4,1X,F10.4,1X,F10.5,1X,F10.5,1X,F10.2)'
      do nr = 1, nretr
        write(recs,fmt=numfmt) nr
        ! exclude vmr values that are anomolously low
        if ((vmr(nr).lt.config%satlowlim).and.(config%satlowlim.gt.0)) then
          write(logid,*) 'WARNING: excluding low satellite value for = ',yyyymmdd, nr 
          cycle
        endif
        cnt = cnt + 1
        ! convert from input unit (e.g. mol/m2) to mole fraction 
        xcpri = cpri(nr)*config%coeffsat/cdryair(nr)
        xcakpri = cakpri(nr)*config%coeffsat/cdryair(nr)
        print*, 'read_obs: xcpri, xcakpri = ',xcpri, xcakpri
        print*, 'read_obs: cpri, cdryair = ',cpri(nr), cdryair(nr)
        print*, 'read_obs: coeffsat = ',config%coeffsat
        ! julian time
        jdate = juldate(idate(nr), itime(nr))
        ! scale retrieval error by factor 
        if (config%saterrfactor.gt.0) then
          vmr_err(nr) = vmr_err(nr)*config%saterrfactor
        endif
        write(200,fmt=rowfmt) recs, idate(nr), itime(nr), jdate, avetime, vmr(nr), vmr_err(nr), &
                              xcpri, xcakpri, cdryair(nr) !, cak_av
      end do

      deallocate( idate, itime, cpri, cakpri, cdryair, vmr, vmr_err )

    else 

      ! check if this file is one to read
      strlen = len_trim(files%suffix)
      string = filelist(nf)
      do i = 1, len_trim(string)
        if ( string(i:(i+strlen)).eq.trim(files%suffix) ) then
          lexist = .true.
        endif
      end do
      if ( .not.lexist ) cycle

      ! ground-based
      ! ------------

      ! check this file belongs to receptor in reclist
      call split(before,"_",recs,sep)
      species = trim(before)
      reclen = len_trim(recs)

      if ( .not.( any(recname(:)(1:reclen).eq.to_upper(recs)) ) ) go to 10

      ! open input file
      open(100,file=trim(files%path_obs)//trim(filelist(nf)),action='read',status='old',iostat=ierr)
      if ( ierr.ne.0 ) then
        write(logid,*) 'WARNING: cannot open: '//trim(files%path_obs)//trim(filelist(nf))
        go to 10
      endif 
      write(logid,*) 'Reading file: '//trim(files%path_obs)//trim(filelist(nf))

      ! read header
      read (100, fmt='(A)', iostat=ierr) header

      ! station coordinates for data selection
      do nr = 1, nrec
        if ( to_upper(recs).eq.recname(nr) ) then
          lon = reclon(nr)
          alt = recalt(nr)
          go to 20
        endif
      end do
20 continue
      if ( lon.eq.0.and.alt.eq.0 ) then 
        write(logid,*) 'ERROR: read_obs: lon and alt are both zero -> check recepter list file'
        stop
      endif

      ! read data
      read_loop: do 
        ! if avetime included in obs file
        write(rowfmt,'(A,I1,A)') '(A',reclen,',1X,I8,1X,I6,1X,F14.6,1X,F10.6,1X,F10.4,1X,F10.4,1X,I4)'
        read(100,fmt=rowfmt,iostat=ierr) recs, yyyymmdd, hhmiss, jdate, avetime, conc, err, num
        ! if avetime not included (backwards compatability with observations prepared before update)
!        avetime = 0d0
!        write(rowfmt,'(A,I1,A)') '(A',reclen,',1X,I8,1X,I6,1X,F14.6,1X,F10.4,1X,F10.4,1X,I4)'
!        read(100,fmt=rowfmt,iostat=ierr) recs, yyyymmdd, hhmiss, jdate, conc, err, num
        if ( ierr.ne.0 ) exit read_loop
        if ( jdate.ge.(juldatef+1d0) ) exit read_loop
        if ( jdate.lt.juldatei ) cycle read_loop
        if ( config%selectobs ) then
          ! select day/night for low/high alt sites
          hh = hhmiss/10000
          hl = hh + int(lon*24./360.)
          if ( hl.lt.0 ) then
            hl = hl + 24
          else if ( hl.ge.24 ) then
            hl = hl - 24
          endif
          if ( alt.le.1000. ) then
            if ( (hl.lt.11).or.(hl.gt.15) ) cycle read_loop
          else
            if ( (hl.gt.3).and.(hl.lt.23) ) cycle read_loop
          endif
        endif
        if ( conc.le.-999. ) cycle read_loop
        measerr = max(config%measerr, err)
        scl_cpri = 0.
        scl_cakpri = 0.
        scl_cdryair = 0.
        cnt = cnt + 1
        ! write to output file
        write(rowfmt,'(A,I1,A)') '(A',recname_len,',1X,I8,1X,I6,1X,F14.6,1X,F10.6,1X,F10.4,1X,F10.4,F10.5,1X,F10.5,1X,F10.2)'
        write(200,fmt=rowfmt) recs, yyyymmdd, hhmiss, jdate, avetime, conc, measerr, scl_cpri, scl_cakpri, scl_cdryair 
      end do read_loop

      ! close input file
      close(100)

10  continue

    endif ! satellite of ground-based

  end do ! loop over files

  ! close output file
  close(200)

  ! count observations
  nobs = nobs + cnt

end subroutine read_obs


